--[[--------------------------------------------------------------------------------------------------
EasyBox

Version: 		004
Author: 		Torben Ratzlaff
Website: 		screenheadmedia.com

Purpose:
Makes it easy to create interface boxes

----------------------------------------------------------------------------------------------------]]

--[[--------------------------------------------------------------------------------------------------
How to use:

Setup box data with EasyBox.newBoxData()
EasyBox.newBoxData() has the following parameter structure:

EasyBox.newBoxData(name, {
	imageSheet = loadedImageSheet, 			--the image sheet containing all neccesary images
	framesCorner 			= {1, 2, 3, 4}, --the frames used for the corners. If a field is left out, the first field is used. If you want to skip a corner use "none".
	framesBorder 			= {1, 2, 3, 4}, --the frames used for the borders. If a field is left out, the first field is used. If you want to skip a corner use "none".
	rotationCorner			= {0, 0, 0, 0}, --the rotation in degrees for each corner. Valid values are 0, 90, 180, 270
	rotationBorder			= {0, 0, 0, 0}, --the rotation in degrees for each border segment.  Valid values are 0, 90, 180, 270
	flipCorner				= {false, false, false, false} --determines if the corner is flipped at the X axes
	flipBorder				= {false, false, false, false} --determines if the border segment is flipped at the X axes
	frameBackground 		= 5, 			--the frame used for the box background, can also be "color" to fill with flat color
	color 					= {1, 1, 1}, 	--a global color, that all garphics are tinted in
	colorCorner 			= {1, 1, 1},	--color for the corners that overrides then global color
	colorBorder 			= {1, 1, 1},	--color for the borders that overrides then global color
	colorBackground 		= {1, 1, 1},	--color for the background that overrides then global color
	alpha 					= 1, 			--a global alpha value that is applied to all graphics
	alphaCorner 			= 1,			--alpha value for the corners that overrides then global alpha
	alphaBorder 			= 1,			--alpha value for the borders that overrides then global alpha
	alphaBackground 		= 1,			--alpha value for the background that overrides then global alpha
	offsetBorder 			= {0, 0, 0, 0}, --determines how many pixels each border is offset (positive = outwards, neagtive = inwards), also effects the border size and the corner positions. If a field is left out, the first field is used.
	offsetCorner 			= {0, 0, 0, 0}, --determines how many pixels each corner is offset (positive = outwards, neagtive = inwards) in x and y direction. If a field is left out, the first field is used.
	offsetContent 			= {0, 0}, 		--the offset of the boxes content origin from the top left corner
})
´--name 			- name of this specific box type


EasyBox.loadBoxDataSet(dataSet)
--dataSet 			- a table containing multiple data tables as used in EasyBox.newBoxData(). Data tables have to contain a "name" field for the box name.


Create new boxs from previously loaded data with EasyBox.newBox()

EasyBox.newBox(name)



Box functions for external use:


Box:remove()	--removes the box


Box:draw(layer, width, height)
--layer 	- the layer the boxs graphics should be inserted into
--width 	- width of the box
--height 	- height of the box
--return 	- returns the boxs grpahic object


Box:erase() --Remove the instances graphics and physics, hide the instace before
--return 	- true if there was a graphic to remove, false if not


Box:show(x, y) --shows the box and sets it's x and y position
--x 		- new x position
--y 		- new y position
--return 	- the boxs graphic object if it exists


Box:hide() --hides the box
--return 	- the boxs graphic object if it exists

Box:setPosition(x, y) --set the position of the box
--x 		- new x position
--y 		- new y position
--return 	- true if the boxs position has been set successfully


Box:getPosition() --returns the boxs current position
--return - x and y position of the box


Box:setSize(width, height)
--width 	- width of the box
--height 	- height of the box
--return 	- true, if the size has been applied to the graphics, false if not


Box:getSize()
--position - determines which width and height values are returned. Possible values are nil (background width and height), "content" (background width and height minus two times content position) and "border" (background width and height plus border offset plus border width)
--return - width, height of the box


Box:getGroup() --returns the boxes group object
--return - returns the boxes display group


Box:getContentGroup() --returns the boxes content group
--return - returns the boxes content group


Box:insert(object, position) --insert a display object into the boxes content group
--object 	- the display object you want to insert
--position 	- the position of the new inserted object (default is on top)

----------------------------------------------------------------------------------------------------]]


local setmetatable = setmetatable
local display_newImage 		= display.newImage
local display_newGroup 		= display.newGroup
local display_newContainer 	= display.newContainer
local display_newRect 		= display.newRect
local display_remove 		= display.remove
local math_floor 	= math.floor
local math_ceil 	= math.ceil
local math_round 	= math.round



local emptyTable = {}

local defaultFrame 			= 1
local defaultRotation		= 0
local defaultFlip			= false
local defaultOffset 		= 0
local defaultWidth 			= 300
local defaultHeight			= 200
local defaultAlpha			= 1



--[[--------------------------------------------------------------------------------------------------
-- setup Box class
----------------------------------------------------------------------------------------------------]]
local Box = {}
local BoxMetaTable = {__index = Box}


--[[--------------------------------------------------------------------------------------------------------
Draw the instances graphics and ad physics
@param arg - a table of additional arguments
@return - the instances graphic handle (graphicObj) if it exists
----------------------------------------------------------------------------------------------------------]]
function Box:draw(layer, width, height)
	local graphicObj = self.graphicObj

	--check if graphicObj already exists
	if not graphicObj then
		--create holder group
		graphicObj = display_newGroup()
		self.graphicObj = graphicObj

		local data = self.data
		local imageSheet = data.imageSheet
		local fillMode = data.fillMode
		local color = data.color
		local alpha = data.alpha or defaultAlpha

		--create background
		local data_frameBackground = data.frameBackground

		if data_frameBackground and data_frameBackground ~= "none" then
			local data_colorBackground = data.colorBackground or color
			local data_alphaBackground = data.alphaBackground or alpha

			local background

			if data_frameBackground == "color" then
				background = display_newRect(0, 0, defaultWidth, defaultHeight)
			else
				background = display_newImage(imageSheet, data_frameBackground)
			end

			if data_colorBackground then
				background:setFillColor(data_colorBackground[1], data_colorBackground[2], data_colorBackground[3])
			end

			self.background = background
			background.alpha = data_alphaBackground

			graphicObj:insert(background)
		end

		--create borders
		local data_framesBorder = data.framesBorder

		if data_framesBorder then
			local data_rotationBorder = data.rotationBorder
			local data_flipBorder = data.flipBorder
			local data_colorBorder = data.colorBorder or color
			local data_alphaBorder = data.alphaBorder or alpha

			local borders = {}
			self.borders = borders

			for i=1, 4 do
				local frame = data_framesBorder[i]

				if frame ~= "none" then
					local border = display_newImage(imageSheet, frame)
					borders[i] = border

					border.rotation = data_rotationBorder[i]
					border.xScale = data_flipBorder[i] and -1 or 1

					border.initialWidth = border.width
					border.initialHeight = border.height

					border.alpha = data_alphaBorder

					if data_colorBorder then
						border:setFillColor(data_colorBorder[1], data_colorBorder[2], data_colorBorder[3])
					end

					graphicObj:insert(border)
				end
			end
		end

		--create corners
		local data_framesCorner = data.framesCorner

		if data_framesCorner then
			local data_rotationCorner = data.rotationCorner
			local data_flipCorner = data.flipCorner
			local data_colorCorner = data.colorCorner or color
			local data_alphaCorner = data.alphaCorner or alpha

			local corners = {}
			self.corners = corners

			for i=1, 4 do
				local frame = data_framesCorner[i]

				if frame ~= "none" then
					local corner = display_newImage(imageSheet, frame)
					corners[i] = corner

					corner.rotation = data_rotationCorner[i]
					corner.xScale = data_flipCorner[i] and -1 or 1

					corner.alpha = data_alphaCorner

					if data_colorCorner then
						corner:setFillColor(data_colorCorner[1], data_colorCorner[2], data_colorCorner[3])
					end

					graphicObj:insert(corner)
				end
			end
		end

		--create contentGroup
		local contentGroup = display_newGroup()
		self.contentGroup = contentGroup
		graphicObj:insert(contentGroup)

		local width 	= width or self.width
		local height 	= height or self.height
		self:setSize(width, height)
	end

	graphicObj.isVisible = false --hide graphic

	if layer then
		layer:insert(graphicObj)
	end

	return graphicObj
end


--[[--------------------------------------------------------------------------------------------------------
Remove the instances graphics and physics, hide the instace before
@return - true if there was a graphic to remove, false if not
----------------------------------------------------------------------------------------------------------]]
function Box:erase()
	local graphicObj = self.graphicObj

	if graphicObj then
		self:hide() --hide the graphic before removing

		display_remove(self.graphicObj)
		self.graphicObj = nil
		self.contentGroup = nil
		self.background = nil
		self.borders = nil
		self.corners = nil

		return true
	end

	return false
end


--[[--------------------------------------------------------------------------------------------------------
Show the instance graphics and position them if they exist

@param x,y - the graphicObj new x and y position
@return - the instances graphic handle (graphicObj) if it exists
----------------------------------------------------------------------------------------------------------]]
function Box:show(x, y)
	local graphicObj = self.graphicObj

	if graphicObj then
		graphicObj.isVisible = true --show graphic

		--apply x and y position
		if x then graphicObj.x = x end
		if y then graphicObj.y = y end
	end

	return graphicObj
end


--[[--------------------------------------------------------------------------------------------------------
Hide the instance graphics and pause it before

@return - the instances graphic handle (graphicObj) if it exists
----------------------------------------------------------------------------------------------------------]]
function Box:hide()
	local graphicObj = self.graphicObj

	if graphicObj then
		graphicObj.isVisible = false
	end

	return graphicObj
end


--[[--------------------------------------------------------------------------------------------------------
Set the graphics x and y position

@param x,y - the graphicObj new x and y position
@return - true if the position has been set successfully
----------------------------------------------------------------------------------------------------------]]
function Box:setPosition(x, y)
	local graphicObj = self.graphicObj

	if graphicObj then
		if x then graphicObj.x = x end
		if y then graphicObj.y = y end

		return true
	end

	return false
end


--[[--------------------------------------------------------------------------------------------------------
Returns the graphics x and y position

@return x,y - the graphicObj x and y position
----------------------------------------------------------------------------------------------------------]]
function Box:getPosition()
	local graphicObj = self.graphicObj

	if graphicObj then
		return graphicObj.x, graphicObj.y
	end
end


--[[--------------------------------------------------------------------------------------------------------
Set the box backgrounds width and height. The size of the borders and corners is added to this values. If the Box has been drawn the new values are applied

@param width,height - the box backgrounds new width and height
@return true - if the size has been applied to the graphics, false if not
----------------------------------------------------------------------------------------------------------]]
function Box:setSize(width, height)
	local graphicObj = self.graphicObj

	local width 	= width or self.width or defaultWidth
	local height 	= height or self.height or defaultHeight

	self.width 		= width
	self.height 	= height

	if graphicObj then
		local data = self.data
		local offsetBorder = data.offsetBorder
		local offsetCorner = data.offsetCorner
		local offsetContent = data.offsetContent

		local halfWidth = width*0.5
		local halfHeight = height*0.5

		--position the content
		local contentGroup = self.contentGroup
		contentGroup.x = -halfWidth + offsetContent[1]
		contentGroup.y = -halfHeight + offsetContent[2]

		local backgroundActualWidth = width
		local backgroundActualHeight = height

		--resize and position background
		local background = self.background
		if background then
			background.width = width
			background.height = height
		end

		--resize and position borders
		local borders = self.borders

		if borders then
			for i=1, 4 do
				local border = borders[i]

				if border then
					local currentBorderOffset 	= offsetBorder[i]
					local nextBorderOffset 		= offsetBorder[i%4 + 1]
					local previousBorderOffset 	= offsetBorder[(i-1)%4] or offsetBorder[4]

					local borderInitialWidth 	= border.initialWidth
					local borderInitialHeight 	= border.initialHeight
					local borderRotation = border.rotation


					
					local borderX, borderY
					local borderWidth, borderHeight

					if i==1 then
						borderWidth = width + nextBorderOffset + previousBorderOffset
						borderHeight = borderInitialHeight

						borderX = (nextBorderOffset - previousBorderOffset)*0.5

						if borderRotation == 90 or borderRotation == 270 then
							borderWidth, borderHeight = borderInitialWidth, borderWidth
							borderY = -borderWidth*0.5 - halfHeight - currentBorderOffset
						else
							borderY = -borderHeight*0.5 - halfHeight - currentBorderOffset
						end
					elseif i==2 then
						borderWidth = borderInitialWidth
						borderHeight = height + nextBorderOffset + previousBorderOffset

						borderY = (nextBorderOffset - previousBorderOffset)*0.5

						if borderRotation == 90 or borderRotation == 270 then
							borderWidth, borderHeight = borderHeight, borderInitialHeight
							borderX = borderHeight*0.5 + halfWidth + currentBorderOffset
						else
							borderX = borderWidth*0.5 + halfWidth + currentBorderOffset
						end

					elseif i==3 then
						borderWidth = width + nextBorderOffset + previousBorderOffset
						borderHeight = borderInitialHeight

						borderX = (previousBorderOffset - nextBorderOffset)*0.5

						if borderRotation == 90 or borderRotation == 270 then
							borderWidth, borderHeight = borderInitialWidth, borderWidth
							borderY = borderWidth*0.5 + halfHeight + currentBorderOffset
						else
							borderY = borderHeight*0.5 + halfHeight + currentBorderOffset
						end
					else
						borderWidth = borderInitialWidth
						borderHeight = height + nextBorderOffset + previousBorderOffset

						borderY = (previousBorderOffset - nextBorderOffset)*0.5

						if borderRotation == 90 or borderRotation == 270 then
							borderWidth, borderHeight = borderHeight, borderInitialHeight
							borderX = -borderHeight*0.5 - halfWidth - currentBorderOffset
						else
							borderX = -borderWidth*0.5 - halfWidth - currentBorderOffset
						end
					end

					border.x = borderX
					border.y = borderY

					--set new width
					border.width = borderWidth
					border.height = borderHeight
				end
			end
		end

		--resize and position corners
		local corners = self.corners

		if corners then
			for i=1, 4 do
				local corner = corners[i]

				if corner then
					local nextBorderOffset 		= offsetBorder[i]
					local previousBorderOffset 	= offsetBorder[(i-1)%4] or offsetBorder[4]

					local currentCornerOffset = offsetCorner[i]

					local defaultCornerX
					local defaultCornerY

					defaultCornerX = (width + corner.width)*0.5
					defaultCornerY = (height + corner.height)*0.5

					--position corner
					local cornerX, cornerY
					if i==1 then
						cornerX = -defaultCornerX - previousBorderOffset - currentCornerOffset
						cornerY = -defaultCornerY - nextBorderOffset - currentCornerOffset
					elseif i==2 then
						cornerX = defaultCornerX + nextBorderOffset + currentCornerOffset
						cornerY = -defaultCornerY - previousBorderOffset - currentCornerOffset
					elseif i==3 then
						cornerX = defaultCornerX + previousBorderOffset + currentCornerOffset
						cornerY = defaultCornerY + nextBorderOffset + currentCornerOffset
					else
						cornerX = -defaultCornerX - nextBorderOffset - currentCornerOffset
						cornerY = defaultCornerY + previousBorderOffset + currentCornerOffset
					end

					corner.x = cornerX
					corner.y = cornerY
				end
			end
		end

		return true
	end

	return false
end


--[[--------------------------------------------------------------------------------------------------------
Returns the box backgrounds width and height

@param position - determines which width and height values are returned. Possible values are nil and "content"
@return width,height - the box backgrounds width and height. if the box asn't been drawn yet these are nil
----------------------------------------------------------------------------------------------------------]]
function Box:getSize(position)
	local width = self.width
	local height = self.height

	if position == "content" then
		local offsetContent = self.data.offsetContent
		width = width - offsetContent[1]*2
		height = height - offsetContent[2]*2
	end

	return width, height
end


--[[--------------------------------------------------------------------------------------------------------
@return display group - the boxs display object
----------------------------------------------------------------------------------------------------------]]
function Box:getGroup()
	return self.graphicObj
end


--[[--------------------------------------------------------------------------------------------------------
@return display group - the boxs content group
----------------------------------------------------------------------------------------------------------]]
function Box:getContentGroup()
	return self.contentGroup
end


--[[--------------------------------------------------------------------------------------------------------
Insert a graphic into the boxs content group hirachie.

@param image - the display object you want to insert
@param position - the graphics position in the boxs content hirachie
----------------------------------------------------------------------------------------------------------]]
function Box:insert(object, position)
	local contentGroup = self.contentGroup

	if not contentGroup then error("[EasyBox] - Box:insert() - Boxs have to be drawn before graphics can be inserted.") end

	if position then
		contentGroup:insert(position, object)
	else
		contentGroup:insert(object)
	end
end


--[[--------------------------------------------------------------------------------------------------
-- setup Module functions
----------------------------------------------------------------------------------------------------]]

local EasyBox = {}
local EasyBoxData = {}

function EasyBox.newBoxData(name, data)
	if EasyBoxData[name] then print("[EasyBox] - EasyBox.newBoxData() - It already exists box data with the name "..name..". Data will be overwritten.") end

	local boxData = {}

	boxData.imageSheet = data.imageSheet


	local data_framesCorner = data.framesCorner
	if data_framesCorner then
		local framesCornerDefault = data_framesCorner[1] or defaultFrame
		boxData.framesCorner = {
			framesCornerDefault,
			data_framesCorner[2] or framesCornerDefault,
			data_framesCorner[3] or framesCornerDefault,
			data_framesCorner[4] or framesCornerDefault,
		}

		local data_rotationCorner = data.rotationCorner or emptyTable
		boxData.rotationCorner = {
			data_rotationCorner[1] or defaultRotation,
			data_rotationCorner[2] or defaultRotation,
			data_rotationCorner[3] or defaultRotation,
			data_rotationCorner[4] or defaultRotation,
		}

		local data_flipCorner = data.flipCorner or emptyTable
		boxData.flipCorner = {
			data_flipCorner[1] or defaultFlip,
			data_flipCorner[2] or defaultFlip,
			data_flipCorner[3] or defaultFlip,
			data_flipCorner[4] or defaultFlip,
		}
	end

	local data_framesBorder = data.framesBorder
	if data_framesBorder then
		local framesBorderDefault = data_framesBorder[1] or defaultFrame
		boxData.framesBorder = {
			framesBorderDefault,
			data_framesBorder[2] or framesBorderDefault,
			data_framesBorder[3] or framesBorderDefault,
			data_framesBorder[4] or framesBorderDefault,
		}

		local data_rotationBorder = data.rotationBorder or emptyTable
		boxData.rotationBorder = {
			data_rotationBorder[1] or defaultRotation,
			data_rotationBorder[2] or defaultRotation,
			data_rotationBorder[3] or defaultRotation,
			data_rotationBorder[4] or defaultRotation,
		}

		local data_flipBorder = data.flipBorder or emptyTable
		boxData.flipBorder = {
			data_flipBorder[1] or defaultFlip,
			data_flipBorder[2] or defaultFlip,
			data_flipBorder[3] or defaultFlip,
			data_flipBorder[4] or defaultFlip,
		}
	end

	boxData.frameBackground = data.frameBackground


	local data_color = data.color
	if data_color then
		boxData.color = {data_color[1], data_color[2], data_color[3]}
	end

	local data_colorCorner = data.colorCorner
	if data_colorCorner then
		boxData.colorCorner = {data_colorCorner[1], data_colorCorner[2], data_colorCorner[3]}
	end

	local data_colorBorder = data.colorBorder
	if data_colorBorder then
		boxData.colorBorder = {data_colorBorder[1], data_colorBorder[2], data_colorBorder[3]}
	end

	local data_colorBackground = data.colorBackground
	if data_colorBackground then
		boxData.colorBackground = {data_colorBackground[1], data_colorBackground[2], data_colorBackground[3]}
	end


	boxData.alpha 			= data.alpha
	boxData.alphaCorner 	= data.alphaCorner
	boxData.alphaBorder 	= data.alphaBorder
	boxData.alphaBackground = data.alphaBackground


	local data_offsetBorder = data.offsetBorder or emptyTable
	local offsetBorderDefault = data_offsetBorder[1] or defaultOffset
	boxData.offsetBorder = {
		offsetBorderDefault,
		data_offsetBorder[2] or offsetBorderDefault,
		data_offsetBorder[3] or offsetBorderDefault,
		data_offsetBorder[4] or offsetBorderDefault,
	}

	local data_offsetCorner = data.offsetCorner or emptyTable
	local offsetCornerDefault = data_offsetCorner[1] or defaultOffset
	boxData.offsetCorner = {
		offsetCornerDefault,
		data_offsetCorner[2] or offsetCornerDefault,
		data_offsetCorner[3] or offsetCornerDefault,
		data_offsetCorner[4] or offsetCornerDefault,
	}

	local data_offsetContent = data.offsetContent or emptyTable
	boxData.offsetContent = {
		data_offsetContent[1] or defaultOffset,
		data_offsetContent[2] or defaultOffset,
	}

	EasyBoxData[name] = boxData
end


local EasyBox_newBoxData = EasyBox.newBoxData
function EasyBox.loadBoxDataSet(dataSet)
	for i=1, #dataSet do
		local boxData = dataSet[i]
		EasyBox_newBoxData(boxData.name, boxData)
	end
end


function EasyBox.newBox(name)
	local data = EasyBoxData[name]

	if not data then error("[EasyBox] - EasyBox.newBox() - No box data for the name "..name) end

	local box = setmetatable({}, BoxMetaTable)
	box.data = data
	
	return box
end


return EasyBox